#include "DistanceTransform.h"

// Offset values to search all 8 pixels in a 3x3 grid
static const int XOFF[] = { 0, 1, 0, -1, -1,  1, 1,-1};
static const int YOFF[] = {-1, 0, 1,  0, -1, -1, 1, 1};
static const float DIST[] = {1,1,1,1,sqrt(2.0),sqrt(2.0),sqrt(2.0),sqrt(2.0)};
static const int DIR = 8; // length of the array XOFF/YOFF/DIST

DistanceTransform::DistanceTransform()
{
    // initialise the variables
    map = NULL;
    DTmap = NULL;
    PFmap = NULL;
    cellState = NULL;

    // The goal and start pointing will be off the screen until the user selects where they are
    goal = XY(-1,-1);
    start = XY(-1,-1);
}

DistanceTransform::~DistanceTransform()
{
    // Clean up
    delete [] map;
    delete [] DTmap;
    delete [] PFmap;
    delete [] cellState;
}

void DistanceTransform::LoadMap(IplImage *image)
{
    // Black is obstacle
    // White is free space

    // If memory has already been allocated dete it
    if(map) {
        delete [] map;
        delete [] DTmap;
        delete [] PFmap;
        delete [] cellState;
    }

    w = image->width;
    h = image->height;

    // Allocate memory
    map = new char[w*h];
    DTmap = new float[w*h];
    PFmap = new float[w*h];
    cellState = new char[w*h];

    // Initialise "map" from the image
    for(int y=0; y < image->height; y++) {
        // Read next row of the image
        unsigned char *buffer = (unsigned char*)&image->imageData[y*image->widthStep];

        for(int x=0; x < image->width; x++) {
            if(buffer[x*3] == 255 && buffer[x*3+1] == 255 && buffer[x*3+2] == 255) // white is freespace
                map[y*w+x] = FREEPSACE;

            if(buffer[x*3] == 0 && buffer[x*3+1] == 0 && buffer[x*3+2] == 0) // black is obstacle
                map[y*w+x] = OBSTACLE;
        }
    }
}

void DistanceTransform::GetNeighbours(int x, int y, vector <XY> &neighbours)
{
    // Search all neighbour cells around (x,y) and append them to the vector neighbours

	for(int i=0; i < DIR; i++) {
		int xx = x + XOFF[i];
		int yy = y + YOFF[i];

		int idx = yy*w + xx;

		if(xx >= 0 && xx < w && yy >= 0 && yy < h)	{ // make sure (xx,yy) is within the image
		    if(cellState[idx] == UNVISITED) {
		        cellState[idx] = VISITED;
                neighbours.push_back(XY(xx, yy));
		    }
		}
	}
}

float DistanceTransform::GetDistance(float *map, int x, int y)
{
    // Returns the lowest cost/distance to get to cell (x,y) by searching its neighbours and finding the one that returns the lowest cost

	float min = 1e9; // initialise minimum cost

	for(int i=0; i < 8; i++) {
		int xx = x + XOFF[i];
		int yy = y + YOFF[i];

		if(xx >= 0 && xx < w && yy >= 0 && yy < h) { // make sure (xx,yy) is within the image
			int idx = yy*w + xx;

			if(cellState[idx] == ASSIGNED) { // this cell has a value assigned to it
			    float d = map[idx] + DIST[i]; // cost function

			    if(d < min)
                    min = d;
			}
		}
	}

	return min;
}

void DistanceTransform::CalcDistanceTransform(bool safePath)
{
    // The results are stored in "DTmap"

    // Make sure there is a valid start/goal point
    if(goal.x < 0 || goal.x >= w || goal.y < 0 || goal.y >= h) // off the image
        return;

    if(start.x < 0 || start.x >= w || start.y < 0 || start.y >= h) // off the image
        return;

    CalcPotentialField();

	// Zero out the memory
	memset(DTmap, 0, sizeof(int)*w*h);
	memset(cellState, 0, sizeof(char)*w*h); // default cell state is UNVISITED

    // Set the goal
	DTmap[goal.y*w + goal.x] = 0; // goal has value of zero (lowest score)
	cellState[goal.y*w + goal.x] = ASSIGNED; // the goal cell has a value ASSIGNED to it

    // Mark the obstacles
    for(int i=0; i < w*h; i++) {
        if(map[i] == OBSTACLE)
            cellState[i] = OBSTACLE;

        // If we're calculating the safe path then mark the cells close to the obstacles as obstacles too
        if(safePath && PFmap[i] < MIN_DISTANCE_FROM_OBSTACLES)
            cellState[i] = OBSTACLE;
    }

    // Get cells around the goal point to visit
    vector <XY> toVisit;
    GetNeighbours(goal.x, goal.y, toVisit);

    maxDTvalue = 0; // initialise maximum distance transform value, for displaying purpose only

	// Sweep out like an ocean wave
	while(toVisit.size()) { // keep processing until there is nothing left to visit
		vector <XY> toVisitNext;

        // Next iteration ...
		while(toVisit.size()) { // keep processing until there is nothing left to visit for this iteration
			XY cur = toVisit.back();
			toVisit.pop_back();

            // Find the lowest cost/distance to this current pixel
			float dist = GetDistance(DTmap, cur.x, cur.y);

            if(dist > maxDTvalue)
                maxDTvalue = dist;

			DTmap[cur.y*w + cur.x] = dist;
            cellState[cur.y*w + cur.x] = ASSIGNED;

			// Get neighbours and append to toVisitNext for next iteration
			GetNeighbours(cur.x, cur.y, toVisitNext);
		}

		toVisit = toVisitNext;
	}
}

void DistanceTransform::CalcPotentialField()
{
    // Calculate the potential field. The value of each cell indicates the distance to its nearest obstacle.
    // This code is very similar to the CalcDistanceTransform function except the goal points are all the obstacles.
    // The results are stored in "PFmap"

	// Initialise
	memset(PFmap, 0, sizeof(int)*w*h);
	memset(cellState, 0, sizeof(char)*w*h); // default cell state is UNVISITED

    // Mark all the obstacles as goal points
    vector <XY> toVisit;

    for(int y=0; y < h; y++) {
        for(int x=0; x < w; x++) {
            if(map[y*w + x] == OBSTACLE) {
                PFmap[y*w + x] = 0; // Set the obstacle as the goal point
                cellState[y*w + x] = ASSIGNED;

                // Get neighbours and append to toVisit
                GetNeighbours(x, y, toVisit);
            }
        }
    }

    maxPFvalue = 0;

	// Sweep out like an ocean wave
	while(toVisit.size()) {
		vector <XY> toVisitNext;

		while(toVisit.size()) {
			XY cur = toVisit.back();
			toVisit.pop_back();

            // Find the shortest distance to this current pixel
			float dist = GetDistance(PFmap, cur.x, cur.y);

            if(dist > maxPFvalue)
                maxPFvalue = dist;

			PFmap[cur.y*w + cur.x] = dist;
            cellState[cur.y*w + cur.x] = ASSIGNED;

			// Get neighbours and append to toVisitNext for next sweep
			GetNeighbours(cur.x, cur.y, toVisitNext);
		}

		toVisit = toVisitNext;
	}
}

void DistanceTransform::FindPath()
{
    // Find a path from (x,y) to the goal
    // Do this using steepest descent
    // The results are stored in "path"

    path.clear();

    // Make sure there is a valid start/goal point
    if(goal.x < 0 || goal.x >= w || goal.y < 0 || goal.y >= h) // off the image
        return;

    if(start.x < 0 || start.x >= w || start.y < 0 || start.y >= h) // off the image
        return;

    if(map[start.y*w + start.x] == OBSTACLE) // Make the user hasn't picked the starting point to be an obstacle
        return;

    // Initialise curVal and prevVal (current and previous cell value)
	float curVal = DTmap[start.y*w + start.x];
	float prevVal;
    XY cur = start;

	do { // At every iteration we find a cell whose value is lower than the previous cell
		path.push_back(cur); // Add the current (x,y) to the path
		prevVal = curVal; // update previous
		XY next;

        // Find the neighbouring cell with the lowest value
		float minVal = 1e8;

		for(int i=0; i < DIR; i++) {
			int xx = cur.x + XOFF[i];
			int yy = cur.y + YOFF[i];

			if(xx >= 0 && xx < w && yy >= 0 && yy < h) { // Make sure (xx,yy) is within the image
				if(DTmap[yy*w + xx] < minVal && cellState[yy*w + xx] != OBSTACLE) {
					minVal = DTmap[yy*w + xx];
					next = XY(xx, yy);
				}
			}
		}

		curVal = minVal;
		cur = next; // next cell to visit
	}  while(curVal < prevVal); // keep looping if current cell value is always lower than the previous cell value
	// The algorithm will end when we reach the goal and there is no lower value that can be found.
}

void DistanceTransform::FindLinearisePath()
{
    FindPath(); // Find the normal distance transform, this will update the local variable "path"
    vector <XY> linearPath; // returned results

    // This algorithm works by finding the longest linear segment that doesn't collide with any obstacle
	for(unsigned int i=0; i < path.size(); i++) {
		XY start = path[i];
		XY end;

		unsigned int j;

		for(j=(i+1); j < path.size(); j++) {
			// Find the current length from path[i] to path[j]
			float dx = path[j].x - path[i].x;
			float dy = path[j].y - path[i].y;
			float mag = dx*dx + dy*dy; // square distance, we avoid square root for efficiency

			if(mag > LINEARISATION_STEP*LINEARISATION_STEP) // path exceeds maximum length
				break;

			if(!IsGoodPath(start, path[j])) // Hit an obstacle, break out
				break;
		}

		end = path[j-1];

		i = j;

        // Append the start and end (X,Y) co-ordinates
		linearPath.push_back(start);
		linearPath.push_back(end);
	}

	path = linearPath;
}

bool DistanceTransform::IsGoodPath(const XY &start, const XY &end)
{
    // Returns true if there is a safe path from start to end eg. no obstacle

    // Calculate the unit vector
	float dx = end.x - start.x;
	float dy = end.y - start.y;
	float mag = sqrt(dx*dx + dy*dy);

	dx /= mag;
	dy /= mag;

	for(int i=0; i <= (int)mag; i++) {
		int x = start.x + dx*i;
		int y = start.y + dy*i;

		int idx = y*w + x;

		if(map[idx] == OBSTACLE)
			return false;
	}

	return true;
}
